__all__ = ["DefaultImageInfoParser", "FasterRCNN", "MaskRCNN"]

from icevision.imports import *
from icevision.parsers.parser import *
from icevision.parsers.mixins import *


class DefaultImageInfoParser(
    Parser,
    FilepathMixin,
    SizeMixin,
    ABC,
):
    """Bundles `Filepath` and `Size` mixins."""


class FasterRCNN(Parser, LabelsMixin, BBoxesMixin, ABC):
    """Parser with required mixins for Faster RCNN."""


class MaskRCNN(FasterRCNN, MasksMixin, IsCrowdsMixin, ABC):
    """Parser with required mixins for Mask RCNN."""



__all__ = ["learner"]

from icevision.imports import *
from icevision.models import efficientdet
from icevision.models.efficientdet.fastai.callbacks import *
from icevision.engines.fastai import *


def learner(
    dls: List[Union[DataLoader, fastai.DataLoader]],
    model: nn.Module,
    param_groups: List[List[nn.Parameter]] = None,
    cbs=None,
    **learner_kwargs,
):
    """Fastai `Learner` adapted for EfficientDet.

    # Arguments
        dls: `Sequence` of `DataLoaders` passed to the `Learner`.
        The first one will be used for training and the second for validation.
        model: The model to train.
        cbs: Optional `Sequence` of callbacks.
        **learner_kwargs: Keyword arguments that will be internally passed to `Learner`.

    # Returns
        A fastai `Learner`.
    """
    cbs = [EfficientDetCallback()] + L(cbs)

    # TODO: Figure out splitter
    learn = adapted_fastai_learner(
        dls=dls,
        model=model,
        cbs=cbs,
        loss_func=efficientdet.loss_fn,
        param_groups=param_groups,
        **learner_kwargs,
    )

    return learn



__all__ = [
    "DataSplitter",
    "SingleSplitSplitter",
    "RandomSplitter",
    "FixedSplitter",
]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *


class DataSplitter(ABC):
    """Base class for all data splitters."""

    def __call__(self, idmap: IDMap):
        return self.split(idmap=idmap)

    @abstractmethod
    def split(self, idmap: IDMap):
        """Splits `ids` into groups.

        # Arguments
            idmap: idmap used for getting ids.
        """
        pass


class SingleSplitSplitter(DataSplitter):
    """Return all items in a single group, without shuffling."""

    def split(self, idmap: IDMap):
        """Puts all `ids` in a single group.

        # Arguments
            idmap: idmap used for getting ids.
        """
        return [idmap.get_ids()]


class RandomSplitter(DataSplitter):
    """Randomly splits items.

    # Arguments
        probs: `Sequence` of probabilities that must sum to one. The length of the
            `Sequence` is the number of groups to to split the items into.
        seed: Internal seed used for shuffling the items. Define this if you need
            reproducible results.

    # Examples

        Split data into three random groups.

        ```python
        idmap = IDMap(["file1", "file2", "file3", "file4"])

        data_splitter = RandomSplitter([0.6, 0.2, 0.2], seed=42)
        splits = data_splitter(idmap)

        np.testing.assert_equal(splits, [[1, 3], [0], [2]])
        ```
    """

    def __init__(self, probs: Sequence[int], seed: int = None):
        self.probs = probs
        self.seed = seed

    def split(self, idmap: IDMap):
        """Randomly splits `ids` based on parameters passed to the constructor of this class.

        # Arguments
            idmap: idmap used for getting ids.
        """
        ids = idmap.get_ids()
        # calculate split indexes
        p = np.array(self.probs) * len(ids)  # convert percentage to absolute
        p = np.ceil(p).astype(int)  # round up, so each split has at least one example
        p[p.argmax()] -= sum(p) - len(ids)  # removes excess from split with most items
        p = np.cumsum(p)

        with np_local_seed(self.seed):
            shuffled = np.random.permutation(list(ids))
        return np.split(shuffled, p.tolist())[:-1]  # last element is always empty


class FixedSplitter(DataSplitter):
    """Split `ids` based on predefined splits.

    # Arguments:
        splits: The predefined splits.

    # Examples

        Split data into three pre-defined groups.
        ```python
        idmap = IDMap(["file1", "file2", "file3", "file4"])
        presplits = [["file4", "file3"], ["file2"], ["file1"]]

        data_splitter = FixedSplitter(presplits)
        splits = data_splitter(idmap=idmap)

        assert splits == [[3, 2], [1], [0]]
        ```
    """

    def __init__(self, splits: Sequence[Sequence[Hashable]]):
        self.splits = splits

    def split(self, idmap: IDMap):
        """Execute the split

        # Arguments
            idmap: idmap used for getting ids.
        """
        return [[idmap.get_name(name) for name in names] for names in self.splits]



__all__ = ["coco", "COCOImageInfoParser", "COCOBBoxParser", "COCOAnnotationParser"]

from icevision.imports import *
from icevision.core import *
from icevision.parsers import *


def coco(
    annotations_file: Union[str, Path],
    img_dir: Union[str, Path],
    mask: bool = True,
) -> ParserInterface:
    annotations_dict = json.loads(Path(annotations_file).read())

    image_info_parser = COCOImageInfoParser(
        infos=annotations_dict["images"], img_dir=img_dir
    )

    _parser_cls = COCOAnnotationParser if mask else COCOBBoxParser
    annotations_parser = _parser_cls(annotations=annotations_dict["annotations"])

    return CombinedParser(image_info_parser, annotations_parser)


class COCOImageInfoParser(DefaultImageInfoParser):
    def __init__(self, infos, img_dir):
        super().__init__()
        self.infos = infos
        self.img_dir = img_dir

    def __iter__(self):
        yield from self.infos

    def __len__(self):
        return len(self.infos)

    def imageid(self, o) -> int:
        return o["id"]

    def filepath(self, o) -> Union[str, Path]:
        return self.img_dir / o["file_name"]

    def height(self, o) -> int:
        return o["height"]

    def width(self, o) -> int:
        return o["width"]


class COCOBBoxParser(FasterRCNN, AreasMixin, IsCrowdsMixin):
    def __init__(self, annotations: list):
        self.annotations = annotations

    def __iter__(self):
        yield from self.annotations

    def __len__(self):
        return len(self.annotations)

    def imageid(self, o) -> int:
        return o["image_id"]

    def labels(self, o) -> List[int]:
        return [o["category_id"]]

    def bboxes(self, o) -> List[BBox]:
        return [BBox.from_xywh(*o["bbox"])]

    def areas(self, o) -> List[float]:
        return [o["area"]]

    def iscrowds(self, o) -> List[bool]:
        return [o["iscrowd"]]


class COCOAnnotationParser(MaskRCNN, COCOBBoxParser):
    def masks(self, o) -> List[MaskArray]:
        seg = o["segmentation"]
        if o["iscrowd"]:
            return [RLE.from_coco(seg["counts"])]
        else:
            return [Polygon(seg)]



__all__ = ["MaskRCNNCallback"]

from icevision.models.rcnn.fastai.callbacks import *
from icevision.models.rcnn.mask_rcnn.prediction import *


class MaskRCNNCallback(RCNNCallback):
    def convert_raw_predictions(self, raw_preds):
        return convert_raw_predictions(
            raw_preds=raw_preds, detection_threshold=0.0, mask_threshold=0.0
        )



__all__ = [
    "resnet18",
    "resnet34",
    "resnet50",
    "resnet101",
    "resnet152",
    "resnext50_32x4d",
    "resnext101_32x8d",
    "wide_resnet50_2",
    "wide_resnet101_2",
    "param_groups",
]

from icevision.imports import *
from icevision.utils import *
from torchvision.models.detection.backbone_utils import resnet_fpn_backbone


def _resnet_fpn(name: str, pretrained: bool = True):
    model = resnet_fpn_backbone(backbone_name=name, pretrained=pretrained)
    model.param_groups = MethodType(param_groups, model)

    return model


def resnet18(pretrained: bool = True):
    return _resnet_fpn("resnet18", pretrained=pretrained)


def resnet34(pretrained: bool = True):
    return _resnet_fpn("resnet34", pretrained=pretrained)


def resnet50(pretrained: bool = True):
    return _resnet_fpn("resnet50", pretrained=pretrained)


def resnet101(pretrained: bool = True):
    return _resnet_fpn("resnet101", pretrained=pretrained)


def resnet152(pretrained: bool = True):
    return _resnet_fpn("resnet152", pretrained=pretrained)


def resnext50_32x4d(pretrained: bool = True):
    return _resnet_fpn("resnext50_32x4d", pretrained=pretrained)


def resnext101_32x8d(pretrained: bool = True):
    return _resnet_fpn("resnext101_32x8d", pretrained=pretrained)


def wide_resnet50_2(pretrained=False):
    return _resnet_fpn("wide_resnet50_2", pretrained=pretrained)


def wide_resnet101_2(pretrained=False):
    return _resnet_fpn("wide_resnet101_2", pretrained=pretrained)


def param_groups(model: nn.Module) -> List[nn.Parameter]:
    body = model.body

    layers = []
    layers += [nn.Sequential(body.conv1, body.bn1)]
    layers += [getattr(body, l) for l in list(body) if l.startswith("layer")]
    layers += [model.fpn]

    _param_groups = [list(layer.parameters()) for layer in layers]
    check_all_model_params_in_groups2(model, _param_groups)

    return _param_groups



__all__ = ["BBox"]

from icevision.imports import *


# TODO: Can pnts and be tuple? Same for properties like xyxy
@dataclass
class BBox:
    """Bounding Box representation.

    Should **not** be instantiated directly, instead use `from_*` methods.
    e.g. `from_xyxy`, `from_xywh`.
    Is able to transform coordinates into different formats,
    e.g. `xyxy`, `xywh`.

    # Examples

    Create from `xywh` format, and get `xyxy` coordinates.
    ```python
    bbox = BBox.from_xywh(1, 1, 4, 4)
    xyxy = bbox.xyxy
    ```
    """

    pnts: List[int]

    def __post_init__(self):
        if self.pnts:
            xl, yu, xr, yb = self.pnts
            self.x, self.y, self.h, self.w = xl, yu, (yb - yu), (xr - xl)
            self.area = self.h * self.w

    def to_tensor(self):
        return tensor(self.xyxy, dtype=torch.float)

    @property
    def xyxy(self):
        return self.pnts

    @property
    def yxyx(self):
        xmin, ymin, xmax, ymax = self.xyxy
        return [ymin, xmin, ymax, xmax]

    @property
    def xywh(self):
        return [self.x, self.y, self.w, self.h]

    def relative_xcycwh(self, img_width: int, img_height: int):
        scale = np.array([img_width, img_height, img_width, img_height])
        x, y, w, h = self.xywh / scale
        xc = x + 0.5 * w
        yc = y + 0.5 * h
        return [xc, yc, w, h]

    @classmethod
    def from_xywh(cls, x, y, w, h):
        return cls([x, y, x + w, y + h])

    @classmethod
    def from_xyxy(cls, xl, yu, xr, yb):
        return cls([xl, yu, xr, yb])

    @classmethod
    def from_relative_xcycwh(cls, xc, yc, bw, bh, img_width, img_height):
        # subtracting 0.5 goes from center to left/upper edge, adding goes to right/bottom
        pnts = [(xc - 0.5 * bw), (yc - 0.5 * bh), (xc + 0.5 * bw), (yc + 0.5 * bh)]
        # convert from relative to absolute coordinates
        scale = np.array([img_width, img_height, img_width, img_height])
        xl, yu, xr, yb = np.around(pnts * scale).astype(int).tolist()
        return cls.from_xyxy(xl, yu, xr, yb)

    @classmethod
    def from_rle(cls, rle, h, w):
        a = np.array(rle.counts, dtype=int)
        a = a.reshape((-1, 2))  # an array of (start, length) pairs
        a[:, 0] -= 1  # `start` is 1-indexed
        y0 = a[:, 0] % h
        y1 = y0 + a[:, 1]
        if np.any(y1 > h):
            # got `y` overrun, meaning that there are a pixels in mask on 0 and shape[0] position
            y0 = 0
            y1 = h
        else:
            y0 = np.min(y0)
            y1 = np.max(y1)
        x0 = a[:, 0] // h
        x1 = (a[:, 0] + a[:, 1]) // h
        x0 = np.min(x0)
        x1 = np.max(x1)
        if x1 > w:
            # just went out of the image dimensions
            raise ValueError(f"invalid RLE or image dimensions: x1={x1} > shape[1]={w}")
        return cls.from_xyxy(x0, y0, x1, y1)



__all__ = ["draw_label", "bbox_polygon", "draw_mask"]

from icevision.imports import *
from matplotlib import patches


def draw_label(ax, x, y, name, color, fontsize=18):
    ax.text(
        x + 1,
        y - 2,
        name,
        fontsize=fontsize,
        color="white",
        va="bottom",
        bbox=dict(facecolor=color, edgecolor=color, pad=2, alpha=0.9),
    )


def bbox_polygon(bbox):
    bx, by, bw, bh = bbox.xywh
    poly = np.array([[bx, by], [bx, by + bh], [bx + bw, by + bh], [bx + bw, by]])
    return patches.Polygon(poly)


def draw_mask(ax, mask, color):
    color_mask = np.ones((*mask.shape, 3)) * color
    ax.imshow(np.dstack((color_mask, mask * 0.5)))
    ax.contour(mask, colors=[color_mask[0, 0, :]], alpha=0.4)



from icevision.imports import *
from icevision.backbones import resnet_fpn
from icevision.models.rcnn.utils import *

from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import (
    maskrcnn_resnet50_fpn,
    MaskRCNN,
    MaskRCNNPredictor,
)


def model(
    num_classes: int,
    backbone: Optional[nn.Module] = None,
    remove_internal_transforms: bool = True,
    **mask_rcnn_kwargs
) -> nn.Module:
    """MaskRCNN model implemented by torchvision.

    # Arguments
        num_classes: Number of classes.
        backbone: Backbone model to use. Defaults to a resnet50_fpn model.
        remove_internal_transforms: The torchvision model internally applies transforms
        like resizing and normalization, but we already do this at the `Dataset` level,
        so it's safe to remove those internal transforms.
        **mask_rcnn_kwargs: Keyword arguments that internally are going to be passed to
        `torchvision.models.detection.mask_rcnn.MaskRCNN`.

    # Return
        A Pytorch `nn.Module`.
    """
    if backbone is None:
        model = maskrcnn_resnet50_fpn(pretrained=True, **mask_rcnn_kwargs)

        in_features_box = model.roi_heads.box_predictor.cls_score.in_features
        model.roi_heads.box_predictor = FastRCNNPredictor(in_features_box, num_classes)

        in_features_mask = model.roi_heads.mask_predictor.conv5_mask.in_channels
        model.roi_heads.mask_predictor = MaskRCNNPredictor(
            in_channels=in_features_mask, dim_reduced=256, num_classes=num_classes
        )

        backbone_param_groups = resnet_fpn.param_groups(model.backbone)
    else:
        model = MaskRCNN(backbone, num_classes=num_classes, **mask_rcnn_kwargs)
        backbone_param_groups = backbone.param_groups()

    patch_param_groups(model=model, backbone_param_groups=backbone_param_groups)

    if remove_internal_transforms:
        remove_internal_model_transforms(model)

    return model



__all__ = ["COCOMetric", "COCOMetricType"]

from icevision.imports import *
from icevision.utils import *
from icevision.data import *
from icevision.metrics.metric import *


class COCOMetricType(Enum):
    """Available options for `COCOMetric`."""

    bbox = "bbox"
    mask = "segm"
    keypoint = "keypoints"


class COCOMetric(Metric):
    """Wrapper around [cocoapi evaluator](https://github.com/cocodataset/cocoapi)

    Calculates average precision.

    # Arguments
        metric_type: Dependent on the task you're solving.
        print_summary: If `True`, prints a table with statistics.
        show_pbar: If `True` shows pbar when preparing the data for evaluation.
    """

    def __init__(
        self,
        metric_type: COCOMetricType = COCOMetricType.bbox,
        print_summary: bool = False,
        show_pbar: bool = False,
    ):
        self.metric_type = metric_type
        self.print_summary = print_summary
        self.show_pbar = show_pbar
        self._records, self._preds = [], []

    def _reset(self):
        self._records.clear()
        self._preds.clear()

    def accumulate(self, records, preds):
        self._records.extend(records)
        self._preds.extend(preds)

    def finalize(self) -> Dict[str, float]:
        with CaptureStdout():
            coco_eval = create_coco_eval(
                records=self._records,
                preds=self._preds,
                metric_type=self.metric_type.value,
                show_pbar=self.show_pbar,
            )
            coco_eval.evaluate()
            coco_eval.accumulate()

        with CaptureStdout(propagate_stdout=self.print_summary):
            coco_eval.summarize()
        # TODO: all results
        mAP = coco_eval.stats[0]
        logs = {"mAP": mAP}

        self._reset()
        return logs



from icevision.parsers.mixins import *
from icevision.parsers.parser import *
from icevision.parsers.combined_parser import *
from icevision.parsers.defaults import *

from icevision.parsers.coco_parser import *
from icevision.parsers.voc_parser import *



__all__ = ["IDMap"]

from icevision.imports import *


class IDMap:
    """
    Works like a dictionary that automatically assign values for new keys.
    """

    def __init__(self, initial_names: Optional[Sequence[Hashable]] = None):
        names = initial_names or []
        self.id2name = OrderedDict((id, name) for id, name in enumerate(names))
        self.name2id = OrderedDict((name, id) for id, name in enumerate(names))

    def get_id(self, id: int) -> Hashable:
        return self.id2name[id]

    def get_name(self, name: Hashable) -> int:
        try:
            id = self.name2id[name]
        except KeyError:
            id = len(self.name2id)
            self.name2id[name] = id
            self.id2name[id] = name

        return id

    def filter_ids(self, ids: List[int]) -> "IDMap":
        idmap = IDMap()
        for id in ids:
            name = self.get_id(id)
            idmap.id2name[id] = name
            idmap.name2id[name] = id

        return idmap

    def get_ids(self) -> List[int]:
        return list(self.id2name.keys())

    def get_names(self) -> List[Hashable]:
        return list(self.name2id.keys())

    def __getitem__(self, imageid):
        return self.get_name(imageid)



__all__ = ["predict", "predict_dl", "convert_raw_predictions"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.models.utils import _predict_dl
from effdet import DetBenchTrain, DetBenchPredict, unwrap_bench


@torch.no_grad()
def predict(
    model: Union[DetBenchTrain, DetBenchPredict],
    batch: Sequence[torch.Tensor],
    detection_threshold: float = 0.5,
    device: Optional[torch.device] = None,
):
    device = device or model_device(model)
    batch = [o.to(device) for o in batch]

    bench = DetBenchPredict(unwrap_bench(model), config=model.config)
    bench = bench.eval().to(device)

    raw_preds = bench(*batch)
    return convert_raw_predictions(raw_preds, detection_threshold=detection_threshold)


def predict_dl(
    model: nn.Module, infer_dl: DataLoader, show_pbar: bool = True, **predict_kwargs
):
    return _predict_dl(
        predict_fn=predict,
        model=model,
        infer_dl=infer_dl,
        show_pbar=show_pbar,
        **predict_kwargs,
    )


def convert_raw_predictions(
    raw_preds: torch.Tensor, detection_threshold: float
) -> List[dict]:
    dets = raw_preds.detach().cpu().numpy()
    preds = []
    for det in dets:
        if detection_threshold > 0:
            scores = det[:, 4]
            keep = scores > detection_threshold
            det = det[keep]

        pred = {
            "scores": det[:, 4],
            "labels": det[:, 5].astype(int),
            "bboxes": [BBox.from_xywh(*xywh) for xywh in det[:, :4]],
        }
        preds.append(pred)

    return preds



__all__ = ["Mask", "MaskArray", "MaskFile", "VocMaskFile", "RLE", "Polygon"]

from icevision.imports import *
from icevision.utils import *
from PIL import Image


class Mask(ABC):
    @abstractmethod
    def to_mask(self, h, w) -> "MaskArray":
        pass

    @abstractmethod
    def to_erle(self, h, w):
        pass


# TODO: Assert shape? (bs, height, width)
@dataclass
class MaskArray(Mask):
    data: np.ndarray

    def __post_init__(self):
        self.data = self.data.astype(np.uint8)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, i):
        return type(self)(self.data[i])

    def to_tensor(self):
        return tensor(self.data, dtype=torch.uint8)

    def to_mask(self, h, w):
        return self

    def to_erle(self, h, w):
        return mask_utils.encode(np.asfortranarray(self.data.transpose(1, 2, 0)))

    def to_coco_rle(self, h, w) -> List[dict]:
        """From https://stackoverflow.com/a/49547872/6772672"""
        assert self.data.shape[1:] == (h, w)
        rles = []
        for mask in self.data:
            counts = []
            flat = itertools.groupby(mask.ravel(order="F"))
            for i, (value, elements) in enumerate(flat):
                if i == 0 and value == 1:
                    counts.append(0)
                counts.append(len(list(elements)))
            rles.append({"counts": counts, "size": (h, w)})
        return rles

    @property
    def shape(self):
        return self.data.shape

    @classmethod
    def from_masks(cls, masks, h, w):
        new = []
        # TODO: Instead of if checks, RLE and Polygon can return with extra dim
        for o in masks:
            m = o.to_mask(h, w).data
            if isinstance(o, (RLE, Polygon)):
                new.append(m[None])
            elif isinstance(o, MaskFile):
                new.append(m)
            elif isinstance(o, MaskArray):
                new.append(m)
            else:
                raise ValueError(f"Segmented type {type(o)} not supported")
        return cls(np.concatenate(new))


@dataclass
class MaskFile(Mask):
    filepath: Union[str, Path]

    def __post_init__(self):
        self.filepath = Path(self.filepath)

    def to_mask(self, h, w):
        mask = open_img(self.filepath, gray=True)
        obj_ids = np.unique(mask)[1:]
        masks = mask == obj_ids[:, None, None]
        return MaskArray(masks)

    def to_coco_rle(self, h, w) -> List[dict]:
        return self.to_mask(h=h, w=w).to_coco_rle(h=h, w=w)

    def to_erle(self, h, w):
        return self.to_mask(h, w).to_erle(h, w)


@dataclass
class VocMaskFile(MaskFile):
    """Extension of `MaskFile` for VOC masks.
    Removes the color pallete and optionally drops void pixels.

    Args:
          drop_void (bool): drops the void pixels, which should have the value 255.
    """

    drop_void: bool = True

    def to_mask(self, h, w) -> MaskArray:
        mask_arr = np.array(Image.open(self.filepath))
        obj_ids = np.unique(mask_arr)[1:]
        masks = mask_arr == obj_ids[:, None, None]

        if self.drop_void:
            masks = masks[:-1, ...]

        return MaskArray(masks)


@dataclass(frozen=True)
class RLE(Mask):
    counts: List[int]

    def to_mask(self, h, w):
        "From https://www.kaggle.com/julienbeaulieu/imaterialist-detectron2"
        mask = np.full(h * w, 0, dtype=np.uint8)
        for start, ones in zip(self.counts[::2], self.counts[1::2]):
            # counting starts on one
            start -= 1
            if ones:
                mask[start : start + ones] = 1
        mask = mask.reshape((h, w), order="F")
        return MaskArray(mask)

    def to_coco(self) -> List[int]:
        coco_counts, total = [], 0
        for start, ones in zip(self.counts[::2], self.counts[1::2]):
            zeros = start - total - 1
            coco_counts.extend([zeros, ones])
            total = start + ones - 1
        # don't include last count if it's zero
        if coco_counts[-1] == 0:
            coco_counts = coco_counts[:-1]
        return coco_counts

    def to_erle(self, h, w):
        return mask_utils.frPyObjects([{"counts": self.counts, "size": [h, w]}], h, w)

    @classmethod
    def from_string(cls, s, sep=" "):
        return cls(lmap(int, s.split(sep)))

    @classmethod
    def from_kaggle(cls, counts):
        """Described [here](https://www.kaggle.com/c/imaterialist-fashion-2020-fgvc7/overview/evaluation)"""
        if len(counts) % 2 != 0:
            raise ValueError("Counts must be divisible by 2")
        return cls(counts)

    @classmethod
    def from_coco(cls, counts):
        """Described [here](https://stackoverflow.com/a/49547872/6772672)"""
        # when counts is odd, round it with 0 ones at the end
        if len(counts) % 2 != 0:
            counts = counts + [0]

        kaggle_counts, total = [], 0
        for zeros, ones in zip(counts[::2], counts[1::2]):
            start = zeros + total + 1
            kaggle_counts.extend([start, ones])
            total += zeros + ones
        return cls.from_kaggle(kaggle_counts)


@dataclass(frozen=True)
class Polygon(Mask):
    points: List[List[int]]

    def to_mask(self, h, w):
        erle = self.to_erle(h=h, w=w)
        mask = mask_utils.decode(erle).sum(axis=-1)  # Sum is for unconnected polygons
        assert mask.max() == 1, "Probable overlap in polygons"
        return MaskArray(mask)

    def to_erle(self, h, w):
        return mask_utils.frPyObjects(self.points, h, w)



__all__ = ["show_annotation"]

from icevision.imports import *
from icevision.utils import *
from icevision.core import *
from icevision.visualize.utils import *
from matplotlib.collections import PatchCollection


# TODO: rename im to img
def show_annotation(
    img,
    labels: List[int] = None,
    bboxes: List[BBox] = None,
    masks=None,
    class_map: ClassMap = None,
    ax: plt.axes = None,
    figsize=None,
    show=False,
) -> None:
    ax = show_img(img=img, ax=ax, figsize=figsize or (10, 10))
    polygons, colors = [], []
    for label, bbox, mask in itertools.zip_longest(
        *[o if o is not None else [] for o in [labels, bboxes, masks]]
    ):
        color = np.random.random(3) * 0.6 + 0.4
        colors.append(color)
        if bbox is not None:
            polygons.append(bbox_polygon(bbox))
        if mask is not None:
            draw_mask(ax, mask.data.copy(), color)
        if label is not None:
            if class_map is not None:
                label = class_map.get_id(label)
            # label position
            if bbox is not None:
                x, y = bbox.x, bbox.y
            elif mask is not None:
                y, x = np.unravel_index(mask.data.argmax(), mask.data.shape)
            else:
                raise ValueError("Can only display labels if bboxes or masks are given")
            draw_label(ax, x=x, y=y, name=label, color=color)

    p = PatchCollection(polygons, facecolor=colors, linewidths=0, alpha=0.3)
    ax.add_collection(p)
    p = PatchCollection(polygons, facecolor="none", edgecolors=colors, linewidths=2)
    ax.add_collection(p)

    if show:
        plt.show()



__all__ = ["FasterRCNNCallback"]

from icevision.models.rcnn.fastai.callbacks import *
from icevision.models.rcnn.faster_rcnn.prediction import *


class FasterRCNNCallback(RCNNCallback):
    def convert_raw_predictions(self, raw_preds):
        return convert_raw_predictions(raw_preds=raw_preds, detection_threshold=0.0)



__all__ = ["Dataset"]

from icevision.imports import *
from icevision.tfms import *
from icevision.data.prepare_record import *


class Dataset:
    """Container for a list of records and transforms.

    Steps each time an item is requested (normally via directly indexing the `Dataset`):
        * Grab a record from the internal list of records.
        * Prepare the record (open the image, open the mask, add metadata).
        * Apply transforms to the record.

    # Arguments
        records: A list of records.
        tfm: Transforms to be applied to each item.
        prepare_record: Function that prepares the record before the transforms are applied.
    """

    def __init__(
        self,
        records: List[dict],
        tfm: Transform = None,
        prepare_record=None,
    ):
        self.records = records
        self.tfm = tfm
        self.prepare_record = prepare_record or default_prepare_record

    def __len__(self):
        return len(self.records)

    def __getitem__(self, i):
        data = self.prepare_record(self.records[i])
        if self.tfm is not None:
            data = self.tfm(data)
        return data

    def __repr__(self):
        return f"<{self.__class__.__name__} with {len(self.records)} items>"

    @classmethod
    def from_images(
        cls, images: Sequence[np.array], tfm: Transform = None, prepare_record=None
    ):
        """Creates a `Dataset` from a list of images.

        # Arguments
            images: `Sequence` of images in memory (numpy arrays).
            tfm: Transforms to be applied to each item.
            prepare_record: Function that prepares the record before the transforms are applied.
        """
        records = []
        for image in images:
            record = {"img": image}
            record["height"], record["width"], _ = image.shape
            records.append(record)

        return cls(records=records, tfm=tfm, prepare_record=prepare_record)



from icevision.utils.utils import *
from icevision.utils.torch_utils import *
from icevision.utils.imageio import *
from icevision.utils.get_files import *
from icevision.utils.download_utils import *
from icevision.utils.data_dir import *
from icevision.utils.capture_stdout import *



__all__ = ["soft_import", "SoftDependencies"]

import importlib
from typing import *


def soft_import(name: str):
    try:
        importlib.import_module(name)
        return True
    except ModuleNotFoundError as e:
        if str(e) != f"No module named '{name}'":
            raise e
        return False


class _SoftDependencies:
    def __init__(self):
        self.fastai = soft_import("fastai")
        self.pytorch_lightning = soft_import("pytorch_lightning")
        self.albumentations = soft_import("albumentations")
        self.effdet = soft_import("effdet")

    def check(self) -> Dict[str, bool]:
        return self.__dict__.copy()


SoftDependencies = _SoftDependencies()



from icevision.imports import *
from icevision.engines.fastai import *
from icevision.models.rcnn.fastai.learner import rcnn_learner
from icevision.models.rcnn.faster_rcnn.fastai.callbacks import *


def learner(
    dls: Sequence[Union[DataLoader, fastai.DataLoader]],
    model: nn.Module,
    cbs: Optional[Sequence[fastai.Callback]] = None,
    **learner_kwargs
) -> fastai.Learner:
    """Fastai `Learner` adapted for Faster RCNN.

    # Arguments
        dls: `Sequence` of `DataLoaders` passed to the `Learner`.
        The first one will be used for training and the second for validation.
        model: The model to train.
        cbs: Optional `Sequence` of callbacks.
        **learner_kwargs: Keyword arguments that will be internally passed to `Learner`.

    # Returns
        A fastai `Learner`.
    """
    cbs = [FasterRCNNCallback()] + L(cbs)
    return rcnn_learner(dls=dls, model=model, cbs=cbs, **learner_kwargs)



__all__ = ["remove_internal_model_transforms", "patch_param_groups"]

from icevision.imports import *
from icevision.utils import *
from torchvision.models.detection.generalized_rcnn import GeneralizedRCNN


def remove_internal_model_transforms(model: GeneralizedRCNN):
    def noop_normalize(image):
        return image

    def noop_resize(image, target):
        return image, target

    model.transform.normalize = noop_normalize
    model.transform.resize = noop_resize


def patch_param_groups(
    model: nn.Module, backbone_param_groups: List[List[nn.Parameter]]
):
    def param_groups(model: nn.Module) -> List[List[nn.Parameter]]:
        head_layers = [model.rpn, model.roi_heads]
        head_param_groups = [list(layer.parameters()) for layer in head_layers]

        _param_groups = backbone_param_groups + head_param_groups
        check_all_model_params_in_groups2(model, _param_groups)

        return _param_groups

    model.param_groups = MethodType(param_groups, model)



__all__ = ["ClassMap"]

from icevision.imports import *


class ClassMap:
    """Utility class for mapping between class name and id."""

    def __init__(self, classes: List[str], background: Optional[int] = 0):
        classes = classes.copy()

        if background is not None:
            if background == -1:
                background = len(classes)
            classes.insert(background, "background")

        self.id2class = classes
        self.class2id = {name: i for i, name in enumerate(classes)}

    def __len__(self):
        return len(self.id2class)

    def get_id(self, id: int) -> str:
        return self.id2class[id]

    def get_name(self, name: str) -> int:
        return self.class2id[name]

    def __repr__(self):
        return f"<ClassMap: {self.class2id.__repr__()}>"



from icevision.tfms.transform import *

import icevision.tfms.batch

# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.albumentations:
    # `import icevision.tfms.albumentations as A` fails in python 3.6
    # https://stackoverflow.com/a/24968941/6772672
    from icevision.tfms import albumentations as A



__all__ = [
    "RecordMixin",
    "ImageidRecordMixin",
    "FilepathRecordMixin",
    "SizeRecordMixin",
    "LabelsRecordMixin",
    "BBoxesRecordMixin",
]
from icevision.imports import *
from .bbox import *


class RecordMixin:
    # TODO: as_dict is only necessary because of backwards compatibility
    def as_dict(self) -> dict:
        return {}


class RecordMetadataMixin:
    pass


class RecordAnnotationMixin:
    pass


class ImageidRecordMixin(RecordMixin):
    def set_imageid(self, imageid: int):
        self.imageid = imageid

    def as_dict(self) -> dict:
        return {"imageid": self.imageid, **super().as_dict()}


class FilepathRecordMixin(RecordMixin):
    def set_filepath(self, filepath: Path):
        self.filepath = filepath

    def as_dict(self) -> dict:
        return {"filepath": self.filepath, **super().as_dict()}


class SizeRecordMixin(RecordMixin):
    def set_image_size(self, width: int, height: int):
        self.width, self.height = width, height

    def as_dict(self) -> dict:
        return {"width": self.width, "height": self.height, **super().as_dict()}


### Annotation parsers ###
class LabelsRecordMixin(RecordMixin):
    def __init__(self):
        super().__init__()
        self.labels: List[int] = []

    def is_valid(self) -> List[bool]:
        return [True for _ in self.labels]

    def add_labels(self, labels: List[int]):
        self.labels.extend(labels)

    def as_dict(self) -> dict:
        return {"labels": self.labels, **super().as_dict()}


class BBoxesRecordMixin:
    def __init__(self):
        super().__init__()
        self.bboxes: List[BBox] = []

    def is_valid(self) -> List[bool]:
        super_valids = super().is_valid()
        valids = [bbox.is_valid(self.width, self.height) for bbox in self.bboxes]

        return valids + super_valids

    def add_bboxes(self, bboxes):
        self.bboxes.extend(bboxes)

    def as_dict(self) -> dict:
        return {"bboxes": self.bboxes, **super().as_dict()}



# import icevision.models.rcnn.backbones as backbones
from icevision.models.rcnn import backbones
from icevision.models.rcnn.loss_fn import *

from icevision.models.rcnn.mask_rcnn.dataloaders import *
from icevision.models.rcnn.mask_rcnn.model import *
from icevision.models.rcnn.mask_rcnn.prediction import *
from icevision.models.rcnn.mask_rcnn.show_results import *

# Soft dependencies
from icevision.soft_dependencies import SoftDependencies

if SoftDependencies.fastai:
    import icevision.models.rcnn.mask_rcnn.fastai

if SoftDependencies.pytorch_lightning:
    import icevision.models.rcnn.mask_rcnn.lightning



from icevision.utils import *
from icevision.core import *
from icevision import parsers
from icevision import tfms
from icevision.data import *
from icevision import backbones
from icevision import models
from icevision.models import *
from icevision.metrics import *
from icevision.visualize import *

import icevision.test_utils as test_utils

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)



__all__ = ["mobilenet", "mobilenet_param_groups"]

from icevision.imports import *
from icevision.utils import *


def mobilenet(pretrained: bool = True):
    model = torchvision.models.mobilenet_v2(pretrained=pretrained)

    features = model.features
    features.out_channels = 1280
    features.param_groups = MethodType(mobilenet_param_groups, features)

    return features


def mobilenet_param_groups(model: nn.Module) -> List[List[nn.Parameter]]:
    layers = []
    layers += [nn.Sequential(*model[0])]
    layers += [nn.Sequential(*model[1:3])]
    layers += [nn.Sequential(*model[3:12])]
    layers += [nn.Sequential(*model[12:])]

    param_groups = [list(layer.parameters()) for layer in layers]
    check_all_model_params_in_groups2(model, param_groups)

    return param_groups



__all__ = [
    "resnet_param_groups",
    "resnet18",
    "resnet34",
    "resnet50",
    "resnet101",
    "resnet152",
    "resnext101_32x8d",
]

from icevision.imports import *
from icevision.utils import *


def _resnet_features(model: nn.Module, out_channels: int):
    # remove last layer (fully-connected)
    modules = list(model.named_children())[:-1]
    features = nn.Sequential(OrderedDict(modules))

    features.out_channels = out_channels
    features.param_groups = MethodType(resnet_param_groups, features)

    return features


def resnet_param_groups(model: nn.Module) -> List[nn.Parameter]:
    layers = []
    layers += [nn.Sequential(model.conv1, model.bn1)]
    layers += [l for name, l in model.named_children() if name.startswith("layer")]

    param_groups = [list(layer.parameters()) for layer in layers]
    check_all_model_params_in_groups2(model, param_groups)

    return param_groups


def resnet18(pretrained: bool = True):
    model = torchvision.models.resnet18(pretrained=pretrained)
    return _resnet_features(model, out_channels=512)


def resnet34(pretrained: bool = True):
    model = torchvision.models.resnet34(pretrained=pretrained)
    return _resnet_features(model, out_channels=512)


def resnet50(pretrained: bool = True):
    model = torchvision.models.resnet50(pretrained=pretrained)
    return _resnet_features(model, out_channels=2048)


def resnet101(pretrained: bool = True):
    model = torchvision.models.resnet101(pretrained=pretrained)
    return _resnet_features(model, out_channels=2048)


def resnet152(pretrained: bool = True):
    model = torchvision.models.resnet152(pretrained=pretrained)
    return _resnet_features(model, out_channels=2048)


def resnext101_32x8d(pretrained: bool = True):
    model = torchvision.models.resnext101_32x8d(pretrained=pretrained)
    return _resnet_features(model, out_channels=2048)
